#version 330

const float PI = 3.14159;

uniform sampler2D albedo;
uniform sampler2D metal;
uniform sampler2D ao;
uniform sampler2D roughness;

in Data {
	vec3 normal;
	vec3 l_dir;
	vec3 pos_cam;
    vec2 texCoord;
} DataIn;

out vec4 colorOut;

float dist_GGX(float alpha, float dot_NH) {

    float denom = PI * pow(dot_NH * dot_NH * (alpha*alpha - 1) + 1, 2);
    if (dot_NH > 0)
        return alpha * alpha / denom;
    else return 0;
}


float geom_GGX_aux(float alpha, float dot_xN, float dot_Hx) {

    float dot_xN_squared = dot_xN * dot_xN;
    float denom = 1 + sqrt(1 + alpha * alpha * (1 - dot_xN_squared)/dot_xN_squared);
    if (dot_Hx/dot_xN > 0)
        return 2 / denom;
    else return 0;

}

float geometric_GGX(float alpha, float dot_HL, float dot_LN, float dot_VN) {

    return geom_GGX_aux(alpha, dot_VN, dot_HL) * geom_GGX_aux(alpha, dot_LN, dot_HL);
}

vec4 fresnel_schlick(vec4 F0, float dot_VH) {

    return F0 + (1.0 - F0) * pow(1.0 - dot_VH, 5.0);
}


void main() {

    // read data from textures
    vec4 diffuse = texture(albedo, DataIn.texCoord);
    float metalness = texture(metal, DataIn.texCoord).r;
    float roughness = 1 - texture(roughness, DataIn.texCoord).r;
    float ao = texture(ao, DataIn.texCoord).r;
    float alpha = 1.0 - roughness;

    diffuse = pow(diffuse, vec4(2.2));

	// normalize both input vectors
	vec3 n = normalize(DataIn.normal);
	vec3 l = DataIn.l_dir;
    
    // compute view vector
	vec3 v = normalize(DataIn.pos_cam);

    // half vector
    vec3 h = normalize(l+v);

	// aux variables to simplify equations

	float dot_NL = clamp(dot(n,l), 0.0, 1.0);
	float dot_NV = clamp(dot(n,v), 0.0, 1.0);
    float dot_HV = clamp(dot(h,v), 0.0, 1.0);
    float dot_NH = clamp(dot(n,h), 0.0, 1.0);
    float dot_HL = clamp(dot(h,l), 0.0, 1.0);

    float d = dist_GGX(alpha, dot_NH);
    float g = geometric_GGX(alpha, dot_HL, dot_NL, dot_NV);

    vec4 F0  = mix(vec4(0.04), diffuse, metalness);
    vec4 f = fresnel_schlick(F0, dot_HV);

    vec4 kd = (1 - f)  ;
    // add small constant to denominator to prevent division by zero
    vec4 specInt = d * f * g / (4 * dot_NL * dot_NV + 0.00001); 

    float intensity = max(0, dot_NL) * 20;

	colorOut = intensity * (specInt  + (1-f)*(1-metalness) * diffuse) + vec4(0.03) * diffuse * ao; 

    // HDR tonemapping
    colorOut = colorOut / (colorOut + vec4(1.0));
    // gamma correct
    colorOut = pow(colorOut, vec4(1.0/2.2)); 

    //colorOut = diffuse;
    
}
